@@ -1,25 +1,40 @@
 #! /usr/bin/perl -w
 #
-# This script creates proper entries for an FDX file as used by the
-# CJK package for LaTeX.
+# This script creates proper entries for an FDX file as used by the CJK
+# package for LaTeX.
 #
 # As a prerequisite, it needs the file
 #
 #   http://partners.adobe.com/public/developer/en/opentype/aglfn13.txt
 #
+# which must be located in the current directory.
+#
 # Call the script as
 #
-#   perl makefdx.pl vertref_afm_file sfd_file fdx_file
+#   perl [-u] makefdx.pl vertref-afm-file sfd-file encoding font-name
 #
-# `vertref_afm_file' is an AFM file as produced by the script `vertref.pe'.
-# The subfont definition file `sfd_file' gives the subfont scheme to get the
-# proper entries in `fdx_file'.
+# `vertref-afm-file' is an AFM file as produced by the script `vertref.pe'.
+# The subfont definition file `sfd-file' gives the subfont scheme to get the
+# proper entries in the FDX file.  `encoding' and `font-name' are TeX font
+# parameters as used by the CJK package; the scripts uses the concatenation
+# of those two values as the name of the FDX file (with suffix `.fdx').
+#
+# The switch `-u' makes the script add a macro to the FDX file (which is
+# used by the CJKutf8.sty) to provide a proper /ToUnicode cmap to pdftex.
 #
 # Note that the created FDX file has to be completed manually.
 #
-# Example:
+# Examples:
 #
-#   perl makefdx.pl bsmiuvr.afm UBig5.sfd c00bsmi.fdx
+#   perl makefdx.pl bsmiuvr.afm UBig5.sfd c00 bsmi
+#
+# The result of this call is the file `c00bsmi.fdx' (you get some warnings
+# because not all glyphs contained in bsmiuvr.afm can be accessed with Big5
+# encoding).
+#
+#   perl makefdx.pl -u bsmiuvr.afm Unicode.sfd c70 bsmi
+#
+# The result of this call is the file `c70bsmi.fdx'.
 #
 # A collection of useful subfont definition files for CJK fonts can be found
 # in the ttf2pk package.
@@ -46,14 +61,22 @@ use strict;
 my $prog = $0;
 $prog =~ s@.*/@@;
 
-if ($#ARGV != 2) {
-  die("usage: $prog vertref_afm_file sfd_file fdx_file\n");
+my $is_unicode = 0;
+if ($ARGV[0] eq "-u") {
+  $is_unicode = 1;
+  shift @ARGV;
+}
+
+if ($#ARGV != 3) {
+  die("usage: $prog [-u] vertref-afm-file sfd-file encoding font-name\n");
 }
 
 my $vertrefafmfile = $ARGV[0];
 my $sfdfile = $ARGV[1];
-my $fdxfile = $ARGV[2];
+my $encoding = $ARGV[2];
+my $fontname = $ARGV[3];
 
+my $fdxfile = $encoding . $fontname . ".fdx";
 
 # Read AGL file.
 
@@ -97,13 +120,14 @@ print("% This is the file $fdxfile of the CJK package
 \\CJKvdef{height}{.88em}
 \\CJKvdef{offset}{.6em}
 
-%\\CJKhdef{offset}{-.12em}
-
 % Uncomment if necessary.
 %\\CJKvdef{norotate}{}
 ");
 
-foreach my $index (0 .. ($#vertref - 1)) {
+my @unicodes;
+my $mapping_count = 0;
+
+foreach my $index (0 .. $#vertref) {
   my $glyphnameref = $vertref[$index];
   my $unicode;
 
@@ -123,8 +147,10 @@ foreach my $index (0 .. ($#vertref - 1)) {
       || ($unicode >= 0xD800 && $unicode <= 0xDFFF)
       || $unicode > 0x10FFFF) {
     print(STDERR "Can't map glyph name \`$glyphnameref' to Unicode.\n");
+    $unicodes[$index] = -1;
     next;
   }
+  $unicodes[$index] = $unicode;
 
   my $sfdentry;
 
@@ -132,17 +158,62 @@ foreach my $index (0 .. ($#vertref - 1)) {
     $sfdentry = $sfd[$unicode];
   }
   else {
+    $unicodes[$index] = -1;
     printf(STDERR "\`%s' (U+%04X) not in subfont encoding\n",
            $glyphnameref, $unicode);
     next;
   }
 
+  $mapping_count++;
+
   print("\\CJKvdef{m/n/$sfdentry}");
   print("{\\def\\CJK\@plane{v}\\selectfont\\CJKsymbol{$index}}\n");
   print("\\CJKvlet{bx/n/$sfdentry}");
   print("{m/n/$sfdentry}\n");
 }
-print("\n\\endinput");
+
+if ($is_unicode) {
+  print("
+\\gdef\\CJK\@cmap\@${fontname}v{
+  \\expandafter\\ifx\\csname CJK\@CMap\@${fontname}v\\endcsname \\relax
+    \\immediate\\pdfobj stream {
+      /CIDInit\\space/ProcSet\\space findresource\\space begin\\space
+        12\\space dict\\space begin\\space
+          begincmap\\space
+            /CIDSystemInfo\\space <<\\space
+              /Registry\\space (TeX)\\space
+              /Ordering\\space (${fontname}v)\\space
+              /Supplement\\space 0\\space >>\\space def\\space
+            /CMapName\\space /TeX-${fontname}v-0\\space def\\space
+            1\\space begincodespacerange\\space
+              <00>\\space <FF>\\space
+            endcodespacerange\\space
+            $mapping_count\\space beginbfchar\\space
+");
+
+  foreach my $index (0 .. $#vertref) {
+    if ($unicodes[$index] != -1) {
+      printf("              <%02X>\\space <%04X>\\space\n",
+             $index, $unicodes[$index]);
+    }
+  }
+
+  print("            endbfchar\\space
+          endcmap\\space
+          CMapName\\space currentdict\\space /CMap\\space defineresource\\space
+          pop\\space
+        end\\space
+      end\\space}
+    \\expandafter\\xdef\\csname CJK\@CMap\@${fontname}v\\endcsname{
+      \\the\\pdflastobj}
+  \\fi
+  \\pdffontattr\\font\@name{
+    /ToUnicode\\space\\csname CJK\@CMap\@${fontname}v\\endcsname\\space 0\\space R}
+}
+
+\\endinput
+");
+}
 
 
 # Read an AGL file.
@@ -243,11 +314,10 @@ sub read_sfdfile {
 }
 
 
-# Read AFM file.
+# Read an AFM file.
 #
 #   $1: Name of the AFM file.
 #   $2: Reference to array which maps glyph indices to glyph names.
-#       `vertical.pe' guarantees that there are no holes in the array.
 
 sub read_afmfile {
   my ($afmfile, $maparray) = @_;
